#!/usr/bin/env python

# ClickCommunicator
#  Manage communication between a controller and Click using either
#  Proclikefs or Socket
#
# Written by Geoff Lawler, Erik Kline
# Copyright (c) 2018 USC/ISI
# Some license here!

import logging
import socket
import os
import os.path
import re
import time

from moa.control import Communicator

LOG = logging.getLogger(__name__)

class ClickCommunicatorException(Exception):
    pass

def click_communicator_except(msg):
    LOG.critical(msg)
    raise ClickCommunicatorException(msg)

class ClickCommunicator(Communicator):
    '''
        Create an empty click configu parser.
    '''
    def __init__(self):
        self._confpath = None
        self._config = {}
        self._socket = None
        self._socket_set = False
        self._parse_time = -1.0
        self._parsed = False
        self._read = None
        self._write = None

    def get_value(self, node, key):
        try:
            return self._config[node][key]
        except ClickCommunicatorException:
            pass

        return None

    def set_value(self, node, key, value):
        '''Set a spefic value. i.e. write this value to the active click configuration.'''
        rval = self._write(node, key, value)
        if not rval:
            LOG.info('Error setting key {} to value {}'.format(key, value))

        return rval

    def get_configuration(self):
        '''
            Return the click configuration as dict[node][key] = ['value1', 'value2', ...]
            data structure. You must call parse() to do the actual parsing before calling this.
        '''
        return self._config

    def parse(self, confpath='/click', force=False):
        '''
        Parse the active click configuration pointed to by the confpath argument.
        '''

        self._confpath = confpath
        if not os.path.exists(self._confpath):
            click_communicator_except('Click config path ({}) not found.'.format(self._confpath))

        if self._parsed and os.path.getmtime(self._confpath) < self._parse_time and not force:
            return

        self._parse_time = time.time()

        if not os.path.isdir(self._confpath):
            self._read = self._read_socket
            self._write = self._write_socket
        else:
            self._read = self._read_file
            self._write = self._write_file

        self._config = {}
        nodes = self._read('list')   # this is in the protocol.

        for node in nodes:
            # 'handlers' is also in the protocol.
            lines = self._read('{}.{}'.format(node, 'handlers'))
            for line in lines:
                try:
                    handler, perm = line.split()
                except ValueError:
                    click_communicator_except(
                        'Error reading {}/{}: {} (from {})'.format(node, handler, line, lines))

                if handler == 'handlers':   # handlers lists itself. skip it.
                    continue

                if perm.startswith('r'):   # only read permissions have readable values...
                    value = self._read('{}.{}'.format(node, handler))
                    if value:
                        if node not in self._config:
                            self._config[node] = {}

                        self._config[node][handler] = {'lines': value, 'permission': perm}
        LOG.info("Parsed Config")
        self._parsed = True

    def _read_socket(self, path):
        '''Send msg to the connected click control socket and return the parsed response.'''
        # for protocol details see: http://read.cs.ucla.edu/click/elements/controlsocket
        # Basic protocol response is like:
        # ZZZ: <msg>
        # DATA NNN
        # ...
        # Where ZZZ is 200 success; not 200 error and NNN is len of DATA in bytes.
        msg = path.replace(os.sep, '.')
        s = self._open_control_socket(self._confpath)
        msg = 'READ ' + msg + '\r\n'
        s.sendall(msg.encode('utf-8'))   # CRLF is expected.
        success, resp = self._read_socket_response(s)
        if not success:
            return False, 'Error reading click socket: {}'.format(resp)

        lines = []
        datasize = int(resp)
        if datasize > 0:
            LOG.debug('reading {} bytes'.format(datasize))
            buf = s.recv(datasize)
            # Not sure why, but "list" puts the number of items first. So remove that
            # if this is a 'list' commanda
            buf = buf.decode('utf-8')
            lines = [t.strip() for t in buf.split('\n') if t]  # remove empty lines and split on \n
            if msg.lower() == 'list':
                lines = lines[1:]

        self._close_control_socket()
        return lines

    def _read_write_status(self, s):
        def _readline(s):
            buf = ''
            while True:
                c = s.recv(1).decode('utf-8')
                if c == '\r':
                    c = s.recv(1).decode('utf-8')
                    if c == '\n':
                        break

                buf += c
            return buf

        resp_line = _readline(s)
        line = re.split(' |-', resp_line)
        if line[0] != '200':
            try:
                err_msg = _readline(s)
            except socket.timeout:
                LOG.debug('socket read timeout')
                err_msg = None

            return False, err_msg

        output = "%s: %s" % (line[0], line[-1])
        return True, output

    def _read_socket_response(self, s):
        '''Read the click response. Return response and amounf of data to read.'''
        def _readline(s):
            buf = ''
            while True:
                c = s.recv(1).decode('utf-8')
                if c == '\r':
                    c = s.recv(1).decode('utf-8')
                    if c == '\n':
                        break

                buf += c
            return buf

        resp_line = _readline(s)
        line = re.split(' |-', resp_line)
        if line[0] != '200':
            try:
                err_msg = _readline(s)
            except socket.timeout:
                LOG.debug('socket read timeout')
                err_msg = None

            return False, err_msg

        _, bytecnt = _readline(s).split()
        return True, bytecnt

    def _read_file(self, subpath):
        path = os.path.join(self._confpath, subpath.replace('.', os.sep))
        lines = []
        try:
            LOG.debug('Reading file {}'.format(path))
            with open(path, 'r') as fd:
                lines = [l.strip() for l in fd.readlines()]
                # Not sure why, but "list" puts the number of items first. So remove that
                # if this is a 'list' command.
                if subpath.lower() == 'list':
                    lines = lines[1:]
        except IOError:
            # not an error. some nodes don't have all keys.
            LOG.debug('path does not exist: {}'.format(path))

        return lines

    def _close_control_socket(self):
        if self._socket_set:
            self._socket.close()

        self._socket_set = False

    def _open_control_socket(self, path):
        if self._socket_set:
            return self._socket
        try:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect(path)
        except Exception as e:
            click_communicator_except(
                'Unable to open click control UNIX socket {}: {}'.format(self._confpath, e))

        s.settimeout(1)  # should be very quick as it's local.

        # read proto header and version.
        buf = s.recv(64).decode('utf-8')
        # should be like "Click::ControlSocket/1.3"
        if not buf.startswith('Click::ControlSocket'):
            click_communicator_except('Bad protocol on click control socket, %s\nexiting.' % buf)

        try:
            _, v = buf.strip().split('/')
            if float(v) < 1.3:
                click_communicator_except('Click control protocol too old at {}'.format(v))
        except (ValueError, TypeError):
            click_communicator_except('Error in click control protocol.')

        self._socket = s
        self._socket_set = True
        return s

    def _write_socket(self, node, key, value):
        s = self._open_control_socket(self._confpath)
        try:
            LOG.debug('writing value "{}" to socket at {}'.format(value, self._confpath))
            # GTL: TODO - check return value of socket send.
            values = value if isinstance(value, list) else [value]
            for v in values:
                cmd = 'write {}.{} {}\r\n'.format(node, key, v)
                LOG.info('writing cmd to socket: {}'.format(cmd))
                cmd = cmd.encode('utf-8')
                s.send(cmd)
        except IOError:
            LOG.warning('Unable to write to socket: {} --> {}'.format(key, value))
            return False

        #s.settimeout(0)
        try:
            success, resp = self._read_write_status(s)
            if not success:
                LOG.info('click responded with error code to write: {}'.format(resp))
                return False   # error in response.

            LOG.info('write response: {}'.format(resp))
        except IOError:
            LOG.info('Unable to read response to write: {} --> {}'.format(key, value))
            # This is OK.
        #s.settimeout(1)

        return True

    def _write_file(self, node, key, value):
        path = os.path.join(self._confpath, node, key)
        try:
            LOG.debug('writing value "{}" to file {}'.format(value, path))
            with open(path, 'w') as fd:
                if isinstance(value, list):
                    for v in value:
                        fd.write(v)
                else:
                    fd.write(value)
        except IOError:
            LOG.warning('Unable to write to file. ({} <-- {})'.format(path, value))
            return False

        return True
