import logging
from typing import List, Mapping
from itertools import chain
import networkx as nx
from . import element, pipeline, obj
from .common import find_input_port, find_output_port

LOG = logging.getLogger(__name__)

# Base Classes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Component:
    ''' A component is a named colleciton of elements that are interconnected
    through a pipeline and can be connected to other components through
    pipelines'''

    def __init__(
        self,
        root: str,
        elements: List[element.Element],
        pipelines: List[pipeline.Pipeline],
        g: nx.Graph,
    ):
        self.root = root
        self.elements = elements
        self.pipelines = pipelines
        self.g = g

        g.nodes[root]['elements'] = elements

    def encode(self) -> List[str]:
        es = self.encode_elements()
        ps = self.encode_pipelines(self.g)
        return es + ps

    def encode_elements(self):
        return list(map(lambda x: x.encode(), self.elements))

    def encode_pipelines(self, g: nx.Graph):
        print("PP ", self.pipelines)
        return list(chain.from_iterable(map(
            lambda x: x.build(g, self.root, self.elements, True), self.pipelines
        )))

components: Mapping[type, type] = {}

# Components ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Interface(Component):

    def __init__(self, name: str, ifx: obj.Interface, g: nx.Graph):

        self.ifx = ifx

        args = [element.Arg("", ifx.id)]

        if name.endswith('input'):
            es = [element.InputInterface(name, args)]
        else:
            es = [element.OutputInterface(name, args)]

        ps = [pipeline.Simple()]

        super().__init__(name, es, ps, g)

components[obj.Interface] = Interface


class XDPFrom(Component):
    def __init__(self, name: str, ifx: obj.Interface, g: nx.Graph):
        self.ifx = ifx

        args = [element.Arg("DEV", ifx.id),
                element.Arg("MODE", ifx.mode)]

        es = [element.XDPFrom(name, args)]
        ps = [pipeline.Simple()]

        super().__init__(name, es, ps, g)

    def encode_elements(self):
        return list(map(lambda x: x.encode(), self.elements))

components[obj.XDPFrom] = XDPFrom

class XDPTo(Component):
    def __init__(self, name: str, ifx: obj.Interface, g: nx.Graph):
        self.ifx = ifx

        args = [element.Arg("DEV", ifx.id),
                element.Arg("MODE", ifx.mode)]

        es = [element.XDPTo(name, args)]
        ps = [pipeline.Simple()]

        super().__init__(name, es, ps, g)

    def encode_elements(self):
        return list(map(lambda x: x.encode(), self.elements))

components[obj.XDPTo] = XDPTo

class Host(Component):

    def __init__(self, name: str, host: obj.Host, g: nx.Graph):

        self.host = host
        args = [
            element.Arg('', host.interface.prefix),
            element.Arg('ETHER', host.interface.prefix),
        ]
        super().__init__(
            name,
            [element.Host(name, args)],
            [pipeline.Simple()],
            g
        )

components[obj.Host] = Host


class Arp(Component):

    def __init__(self, name: str, ifx: obj.Arp, g: nx.Graph):

        self.arp = ifx
        arp_args = [
            element.Arg(self.arp.interface.prefix, self.arp.interface.mac)
        ]
        cls_args = {
            element.Arg("12/0806", "20/002"),
            element.Arg("", "-")
        }
        super().__init__(
            name,
            [
                element.Classifier(name, arp_args),
                element.Arp(name, cls_args)
            ],
            [pipeline.Arp()],
            g
        )

components[obj.Arp] = Arp


class VxlanDemux(Component):

    def __init__(self, name: str, dmx: obj.VxlanDemux, g: nx.Graph):

        self.dmx = dmx

        args: List[element.Arg] = []

        for vni in dmx.vnis:

            next_name = 'vxlan_%s_decap'%vni
            vxlan_passthru = list(g.neighbors(next_name))[0]
            dmx.out_order.append(vxlan_passthru)

            args.append(element.Arg("VXLAN", vni))

        for nbr in g.neighbors(name):
            if nbr.endswith('_tap'):
                dmx.out_order.append(nbr)
                break


        super().__init__(
            name,
            [element.VxlanDemux(name, args)],
            [pipeline.Demux()],
            g
        )

components[obj.VxlanDemux] = VxlanDemux

class Queue(Component):

    def __init__(self, name: str, q: obj.Queue, g: nx.Graph):

        self.q = q

        args: List[element.Arg] = [
            element.Arg("", q.size)
        ]

        super().__init__(
            name,
            [element.Queue(name, args)],
            [pipeline.Simple()],
            g
        )

components[obj.Queue] = Queue

class Vxlan(Component):

    def __init__(self, name: str, vxlan: obj.Vxlan, g: nx.Graph):

        self.vxlan = vxlan

        pipelines = []
        if vxlan.encap:
            pipelines.append(pipeline.Chain())

        vxlan_args = [
            element.Arg("SRC_MAC", vxlan.src_mac),
            element.Arg("DST_MAC", vxlan.dst_mac),
            element.Arg("SRC_IP", vxlan.src_ip),
            element.Arg("DST_IP", vxlan.dst_ip),
            element.Arg("VXLAN_ID", vxlan.vni),
        ]

        check_args = [
            element.Arg("", 14)
        ]

        elements = [
            element.VxlanEncap(name, vxlan_args),
            element.CheckIpHeader(name, check_args)
        ]

        super().__init__(name, elements, pipelines, g)

components[obj.Vxlan] = Vxlan

class Mpl(Component):

    def __init__(self, name: str, mpl: obj.Mpl, g: nx.Graph):

        self.mpl = mpl
        self.in_port_counter = 0

        if mpl.mpl_type == 'broadcast':
            elements = [element.Tee(name)]


        elif mpl.mpl_type in ['switched', 'wireless']:

            args: List[element.Arg] = []

            print("INPUT: ", mpl.inputs)
            for key in mpl.inputs:
                mpl.in_order.append(mpl.inputs[key]+"_input")
                #mpl.out_order.append(mpl.outputs[key])

                if mpl.args.macs:
                    args.append(element.Arg("MAC", mpl.args.macs[key]))


            if mpl.mpl_type == 'switched':
                elements = [element.SwitchedMpl(name)]
            else:
                elements = [element.WirelessMpl(name, args)]

        if mpl.mpl_type in ['broadcast', 'switched', 'wireless']:
            pipelines = [pipeline.Demux()]
        else:
            pipelines = []

        super().__init__(name, elements, pipelines, g)

components[obj.Mpl] = Mpl

class Wireless(Component):

    def __init__(self, name: str, wireless: obj.Wireless, g: nx.Graph):

        self.wireless = wireless

        topo_file = "/var/moa/wireless/emulation_topo_" + wireless.ctx.eid + ".topo"
        mob_file = "/var/moa/wireless/emulation_mob_" + wireless.ctx.eid + ".mob"

        with open(topo_file, 'w') as fh:

            fh.write("%s\n" % wireless.topo.boundaries.encode())

            for n in wireless.topo.nodes:
                fh.write("%s\n", n.encode())

        if wireless.mobility:
            with open(mob_file, 'w') as fh:
                fh.write(wireless.mobility.encode())

        args = [
            element.Arg("INTERVAL", wireless.interval),
            element.Arg("TOPOFILE", topo_file),
            element.Arg("MOBFILE", mob_file),
            element.Arg("MOBMODEL", wireless.mobility_model),
            element.Arg("MOBILITY_ACTIVE", wireless.use_mobility)
        ]

        super().__init__(
            name,
            [element.WirelessManager(name, args)],
            [],
            g
        )

components[obj.Wireless] = Wireless

class Link(Component):

    def __init__(self, name: str, link: obj.Link, g: nx.Graph):

        self.link = link

        # wireless elements
        collision = None
        if link.args.wireless:

            neighbor = list(g.neighbors(name))[0]

            downstream_port = find_input_port(g, name, neighbor, True)
            upstream_port = -1

            if self.g.in_degree(name) == 1:
                neighbor = list(self.g.in_edges(name))[0][0]
                upstream_port = find_output_port(self.g, name, neighbor, True)

            if link.args.wireless.collision_handler == 'CSMA':
                collision = element.Csma(name, [
                    element.Arg("", downstream_port)
                ])

            elif link.args.wireless.collision_handler == 'tx_collision':
                collision = element.TxCollision(name, [
                    element.Arg("", downstream_port)
                ])

            elif link.args.wireless.collision_handler == 'rx_collision':
                collision = element.RxCollision(name, [
                    element.Arg("", upstream_port)
                ])

            else:
                LOG.warning("unknown handler '%s'", link.args.wireless.collission_handler)


        # standard elements
        elements = []
        if link.args.use_codel:
            elements.append(element.SetTimestamp(name, [element.Arg("FIRST", "true")]))

        elements.append(element.ThreadSafeQueue(name, [element.Arg("", int(1e6))]))

        if link.args.use_codel:
            elements.append(element.Codel(name))

        if link.args.loss:
            elements.append(element.RandomSample(name, [
                element.Arg("DROP", link.args.loss),
            ]))

        if collision:
            elements.append(collision)

        elements.append(element.LinkUnqueue(name, [
            element.Arg("LATENCY", link.args.delay),
            element.Arg("BANDWIDTH", link.args.bandwidth)
        ]))

        super().__init__(name, elements, [pipeline.Chain()], g)

components[obj.Unilink] = Link

class Mime(Component):

    def __init__(self, name: str, mime: obj.Mime, g: nx.Graph):

        self.mime = mime

        args: List[element.Arg] = [
            element.Arg("IP", mime.ip),
            element.Arg("MAC", mime.mac)
        ]

        for i, f in enumerate(mime.flows):
            args.append(element.Arg("COMPUTE%d"%i, f.compute))

            if f.data:
                args.append(element.Arg("DATA%d"%i, f.data))

            if f.dst:
                args.append(element.Arg("DST%d"%i, f.dst))

            if f.mac:
                args.append(element.Arg("MAC%d"%i, f.mac))

            if f.maxpkt:
                args.append(element.Arg("MAXPKT%d"%i, f.maxpkt))

            if f.minpkt:
                args.append(element.Arg("MINPKT%d"%i, f.minpkt))

        if mime.stop:
            args.append(element.Arg("STOP", "true"))

        elements = [element.Mime(name, args)]
        pipelines = [pipeline.Simple()]

        super().__init__(name, elements, pipelines, g)


components[obj.Mime] = Mime

class Router(Component):

    def __init__(self, name: str, router: obj.Router, g: nx.Graph):

        self.router = router

        args: List[element.Arg] = [
            element.Arg("", x.arg()) for x in router.routes
        ]

        for key in router.inputs:
            router.in_order.append(router.inputs[key])
            router.out_order.append(router.outputs[key])

        elements = [element.Router(name, args)]
        pipelines = [pipeline.Demux()]

        super().__init__(name, elements, pipelines, g)

components[obj.Router] = Router
