from dataclasses import dataclass
from typing import List, Any

@dataclass
class Arg:
    name: str
    value: Any

    def encode(self):
        if self.name:
            return "{} {}".format(self.name, self.value)

        return "{}".format(self.value)

class Element:

    def __init__(self, name: str, click_type: str, args: List[Arg], suffix=""):
        if suffix:
            self.name = name + "_" + suffix
        else:
            self.name = name
        self.click_type = click_type
        if args is None:
            self.args = []
        else:
            self.args = args
        self.suffix = suffix

    def encode_args(self):
        args = ""
        for a in self.args:
            args += a.encode()+', '

        return args.rstrip(', ')

    def encode(self) -> str:
        return "%s :: %s(%s);" % (self.name, self.click_type, self.encode_args())

# Elements ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

## Interfaces

class InputInterface(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "FromDevice", args)

class OutputInterface(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "ToDevice", args)


class XDPFrom(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "XDPFromDevice", args)

class XDPTo(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "XDPToDevice", args)

## Generic packet processing

class VxlanDemux(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "VXLANDemux", args, suffix="vxlan_decap")

class Host(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "KernelTap", args, suffix="host")

class Arp(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "ARPResponder", args, suffix="arp")

class Classifier(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "Classifier", args, suffix="classifier")

class Queue(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "Queue", args)

class VxlanEncap(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "VXLANEncap", args, suffix="vxlan_encap")

class CheckIpHeader(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "CheckIPHeader", args, suffix="check")

class ThreadSafeQueue(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "ThreadSafeQueue", args, suffix="ts_queue")

class SetTimestamp(Element):

    def __init__(self, name:str, args: List[Arg] = None):
        super().__init__(name, "SetTimestamp", args, suffix="set_ts")

class Codel(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "CoDel", args, suffix="codel")

class LinkUnqueue(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "LinkUnqueue", args, suffix="bwdelay")

class RandomSample(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "RandomSample", args, suffix="random_sample")

class SwitchedMpl(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "EtherSwitch", args, suffix="smpl")

class WirelessMpl(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "ReachabilitySwitch", args, suffix="wmpl")

class Tee(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "Tee", args, suffix="tee")

## Wireless

class WirelessManager(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "WirelessManager", args, suffix="wirelesss_manager")

class Csma(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "SimpleCSMA", args, suffix="csma")

class TxCollision(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "CheckTxCollision", args, suffix="tx_collision")

class RxCollision(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "CheckRxCollision", args, suffix="rx_collision")


## Device emulators

class Mime(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "Mime", args, suffix="mime")

class Router(Element):

    def __init__(self, name: str, args: List[Arg] = None):
        super().__init__(name, "MultipathRouter", args, suffix="router")
