#!/usr/bin/env python3

# pylint apparently wants all variables defined in global scopes to be in all
# caps, yeah .....
# pylint: disable=invalid-name

# This script parses the output of `ifconfig` and finds interfaces that begin
# with the "vlan" prefix.  It outputs the vlan ID, IP address, and hardware
# address for each interface.

import re
import subprocess
import json
import ipaddress

PCIADDRS = {}

# use the output of 'ethtool -i' to get the pci address of the ethernet port/interface
def get_pci_address(devname):
    if devname not in PCIADDRS:
        info = {}
        lines = subprocess.check_output(
            ['ethtool', '-i', devname]).decode('utf-8').splitlines()

        for l in lines:
            k, v = l.split(':', maxsplit=1)
            info[k] = v.strip()
        PCIADDRS[devname] = info['bus-info']

    return PCIADDRS[devname]

def get_linkname(ip):
    "Return the NS link name associated with the given IP address"
    # the first hostname in the list is <hostname>-<linkname>
    with open('/etc/hosts') as f:
        for row in f:
            if row.startswith(ip):
                host = row.split()[1]
                return host.split('-')[1]
    return None

# vlan360   Link encap:Ethernet  HWaddr d8:9d:67:ee:ff:54
vlan_re = re.compile(r'vlan(?P<vlan>\d+).*HWaddr (?P<hwaddr>[a-f0-9:]+)')
addr_re = re.compile(r'inet addr:(?P<addr>[0-9.]+).*Mask:(?P<mask>[0-9.]+)')
eth_re = re.compile(r'(?P<eth>eth\d+).*HWaddr (?P<hwaddr>[a-f0-9:]+)')

# map of hwaddr to physical interface
hwmap = {}

realdevs = {}

# maps physical device to DPDK port number
ports = {}

out = subprocess.check_output(['ifconfig'])
state = 0
ifaces = []
port = 0
for line in out.decode('utf-8').splitlines():
    if state == 0:
        m = vlan_re.match(line)
        if m:
            state = 1
            vlan = int(m.group('vlan'))
            device = 'vlan%d' % vlan
            hwaddr = m.group('hwaddr')
            realdev = hwmap[hwaddr]
            kw = {
                'vlan': vlan,
                'interface': device,
                'hwaddr': hwaddr,
                'realdev': realdev,
                'port': ports[realdev],
            }
            realdevs[hwaddr]['vlans'].append(vlan)
        m = eth_re.match(line)
        if m:
            hwaddr = m.group('hwaddr')
            device = m.group('eth')
            if device == 'eth0':
                continue # skip control net
            pciaddr = get_pci_address(device)
            hwmap[hwaddr] = device
            state = 1
            kw = {
                'interface': device,
                'hwaddr': hwaddr,
                'pciaddr': pciaddr,
                'vlans': [],
                'port': port,
            }
            realdevs[hwaddr] = kw
            ports[device] = port
            port += 1
    elif state == 1:
        if not line:
            state = 0
            ifaces.append(kw)
        else:
            m = addr_re.search(line)
            if m:
                addr = m.group('addr')
                mask = m.group('mask')
                ipnet = ipaddress.IPv4Network('%s/%s' % (addr, mask), strict=False)
                cidr = str(ipnet)

                # How DPDK enumerates the physical interfaces to its concept of
                # port numbers is a mystery. For now this code assumes that
                # there is a single multi-port card in an experiment node, and
                # that they are allocated in order by the PCI address (lowest
                # first). port = int(pciaddr[-1])

                kw.update({
                    'ip': addr,
                    'netmask': mask,
                    'cidr': cidr,
                    'link': get_linkname(addr),
                })

print(json.dumps(ifaces))
