import copy
import networkx as nx
import matplotlib
import matplotlib.pyplot as plt
from . import obj
from . import common
matplotlib.use('Agg')

class MoaGraph():
    def __init__(self):
        self.g = nx.DiGraph()   # The actual internal graph Model
        self.roots = []
        self.unconfigured = []

        self.factory = {
            obj.Link: self.add_link,
            obj.Unilink: self.add_unilink,
            obj.Mpl: self.add_mpl,
            obj.Interface: self.add_interface,
            obj.XDPFrom: self.add_interface,
            obj.XDPTo: self.add_interface,
        }

    def add_object(self, x):
        t = type(x)
        f = self.factory.get(t)
        if f is not None:
            return f(x)

        self.add_node(x.id, t.__name__.lower(), x)
        if x.root:
            self.roots.append(x.id)
        return True

    def serialize(self):
        serial_obj = {'roots': self.roots, 'unconfigured': self.unconfigured,
                      'g': nx.node_link_data(self.g)}
        return serial_obj

    def deserialize(self, serial_obj):
        self.roots = serial_obj['roots']
        self.unconfigured = serial_obj['unconfigured']
        self.g = nx.node_link_graph(serial_obj['g'])

    def get_moa_graph_types(self):
        return list(obj.objects.values())

    def get_roots(self):
        return self.roots

    def get_graph(self):
        return self.g

    def add_node(self, name, ntype, info):
        if not (name and ntype and info):
            raise Exception()
        self.g.add_node(name, type=ntype, info=info, visited=False)

    def add_link(self, link: obj.Link):
        status = self.add_unilink(link.left_to_right)
        if status:
            status = self.add_unilink(link.right_to_left)

        return status

    def add_unilink(self, unilink):
        link_out = self.find_io_from_io_point(unilink.output, 'output')
        link_name = unilink.id
        self.add_node(link_name, 'link', unilink)
        for iputs in unilink.inputs:
            link_in = self.find_io_from_io_point(iputs, 'input')
            self.g.add_edge(link_in, link_name)

        self.g.add_edge(link_name, link_out)

        return True

    def insert_queues(self):
        todo = {}
        for node in self.g:
            props = self.g.nodes[node]
            print(node)
            print(props)
            info = props['info']
            if common.is_virtual_output(node, info):
                todo[node] = props['info']
            if hasattr(info, "inQ"):
                if info.inQ:
                    todo[node] = props['info']

        for node, info in todo.items():
            self.add_input_queue(node, info)

    def fix_unconfigured_edges(self):
        new_list = []
        unconfigured = copy.copy(self.unconfigured)
        for node in unconfigured:
            # In edges are outputs from the input node
            correct_name = node.split('_uc')[0]

            in_edges = copy.deepcopy(self.g.in_edges(node))
            for edge in in_edges:
                neighbor = edge[0]
                io_point = self.find_io_from_io_point(correct_name, 'output')
                self.g.remove_edge(edge[0], edge[1])
                self.g.add_edge(neighbor, io_point)

            out_edges = copy.deepcopy(self.g.out_edges(node))
            for edge in out_edges:
                neighbor = edge[1]
                io_point = self.find_io_from_io_point(correct_name, 'input')
                self.g.remove_edge(edge[0], edge[1])
                self.g.add_edge(io_point, neighbor)

            if self.g.degree(node) <= 0:
                self.g.remove_node(node)
            else:
                new_list.append(node)

        self.unconfigured = new_list
        return len(self.unconfigured)

    def convert_string_to_object(self, io_str):
        name = ""
        if io_str.startswith('vxlan'):
            # If we have a vxlan, we must have a encap element
            name = io_str + "_encap"
        elif io_str.startswith('mpl'):
            name = io_str
        elif io_str.startswith('link'):
            name = io_str
        else:
            # Everything else is an interface
            name = io_str + "_input"

        ret = None

        if name in self.g:
            ret = self.g.nodes[name]['info']

        #TODO(ry) hack for non-implicit names
        elif io_str in self.g:
            ret = self.g.nodes[io_str]['info']

        return ret

    def add_unconfigured(self, name, data):
        base_uc_name = name + '_uc'
        uc_name = base_uc_name
        count = 1
        while uc_name in self.g:
            uc_name += '_%d' % count
            count += 1

        self.add_node(uc_name, 'Unconfigured', data)
        self.unconfigured.append(uc_name)
        return uc_name

    def find_io_from_io_point(self, io_point, io_type):
        if io_type == 'input':
            iface_trailer = '_input'
            vxlan_trailer = '_decap'
        else:
            iface_trailer = '_output'
            vxlan_trailer = '_encap'

        io_obj = ''
        if isinstance(io_point, str):
            new_io_point = self.convert_string_to_object(io_point)
            if not new_io_point:
                # References a non-existent IO point.  We'll search for
                uc_name = self.add_unconfigured(io_point, io_type)
                return uc_name

            io_point = new_io_point

        if isinstance(io_point, (obj.Interface, obj.XDPFrom, obj.XDPTo)):
            io_obj = io_point.id + iface_trailer

            # Adding interface if not found, covers both sides.
            if io_obj not in self.g.nodes:
                self.add_interface(io_point)

        elif isinstance(io_point, obj.Vxlan):
            io_obj = 'vxlan_' + io_point.id + vxlan_trailer
            if io_obj not in self.g.nodes:
                self.add_vxlan(io_point)
        elif isinstance(io_point, obj.Object):
            io_obj = io_point.id
        else:
            io_obj = None
        return io_obj

    def add_interface(self, interface):
        iface = interface.id

        t = type(interface)

        our_type = 'interface'
        if t == obj.XDPFrom:
            our_type = 'xdpfrom'
            iface = iface + '_input'
        if t == obj.XDPTo:
            our_type = 'xdpto'
            iface = iface + '_output'

        if iface in self.g.nodes:
            #Possibly flag error here?
            #It could be fine that we're trying to add an interface twice in
            #the XDP Case
            return our_type in ('xdpfrom', 'xdpto')

        self.add_node(iface, our_type, interface)

        self.roots.append(iface)

        if t in (obj.XDPFrom, obj.XDPTo):
            return True

        return self.add_host_handling(interface)

    def add_host_handling(self, interface):
        tap_name = interface.id + '_tap'
        iface_output = interface.id + '_output'

        if tap_name in self.g.nodes:
            return False

        info = {'interface': interface}

        self.add_node(tap_name, 'host', info)
        self.g.add_edge(tap_name, iface_output)

        return True

    def add_vxlan_demux(self, vxlans, interface_obj=None):
        # Assume all VXLANs part of same interface, probably a bad idea
        if interface_obj is None:
            interface_obj = vxlans[0]['interface']
        vxlan_demux = interface_obj['id'] + '_vxlan_demux'
        tap_name = interface_obj['id'] + '_tap'

        if vxlan_demux in self.g:
            return False

        self.add_node(vxlan_demux, 'vxlan_demux',
                     vxlans)

        arp = interface_obj['id'] + '_arp_handler'

        if arp not in self.g.nodes:
            self.add_arp_handling(interface_obj)

        self.g.add_edge(arp, vxlan_demux)
        self.g.add_edge(vxlan_demux, tap_name)
        return True

    def add_arp_handling(self, interface):
        arp = interface.id + '_arp_handler'

        if arp in self.g:
            return False

        self.add_node(arp, 'arp', interface)

        iface_input = interface.id + '_input'
        iface_output = interface.id + '_output'

        if iface_input not in self.g.nodes:
            self.add_interface(interface)

        self.g.add_edge(iface_input, arp)
        self.g.add_edge(arp, iface_output)

        return True

    def add_input_queue(self, name, interface):
        queue = '%s_queue'%name
        self.add_node(queue, 'queue', obj.Queue(obj.Context(""), name))

        in_edges = self.g.in_edges(name)
        to_remove = []
        to_add = []
        for edge in in_edges:
            if self.g.nodes[edge[0]]['type'] != 'link':
                continue
            to_remove.append((edge[0], edge[1]))
            to_add.append((edge[0], queue))

        for a, b in to_remove:
            self.g.remove_edge(a, b)

        for a, b in to_add:
            self.g.add_edge(a, b)

        self.g.add_edge(queue, name)

    def add_vxlan_to_demux(self, vxlan_demux, vxlan):
        if vxlan not in self.g.nodes[vxlan_demux]['info']:
            self.g.nodes[vxlan_demux]['info'].append(vxlan)
        return True

    def add_vxlan(self, vxlan):
        vxlan_encap = 'vxlan_%s_encap' % vxlan['id']
        vxlan_decap = 'vxlan_%s_decap' % vxlan['id']

        interface = vxlan['interface']

        if interface is None:
            # Interface needs to be defined in the vxlan
            return False

        if isinstance(interface, str):
            tmp_iface = interface + '_input'
            interface = self.g.node[tmp_iface]['info']
            if not interface:
                return False

            vxlan['interface'] = interface

        if vxlan_encap in self.g:
            return False

        vxlan_demux = interface.id + '_vxlan_demux'
        iface_output = interface.id + '_output'

        if iface_output not in self.g.nodes:
            self.add_interface(interface)

        if vxlan_demux not in self.g.nodes:
            self.add_vxlan_demux([vxlan], interface)
        else:
            self.add_vxlan_to_demux(vxlan_demux, vxlan)

        self.add_node(vxlan_encap, 'vxlan', vxlan)
        self.add_node(vxlan_decap, 'vxlan', vxlan)

        # Encap -> Output
        self.g.add_edge(vxlan_encap, iface_output)

        # Input -> VXLAN Decap
        self.g.add_edge(vxlan_demux, vxlan_decap)

        return True

    def add_mpl(self, mpl):
        if mpl.mpl_type == 'broadcast':
            self.add_broadcast_mpl(mpl)
        elif mpl.mpl_type == 'switched' or mpl.mpl_type == 'wireless':
            self.add_switched_mpl(mpl)
            if mpl.mpl_type == 'wireless':
                if 'wireless' in mpl.args:
                    self.add_wireless(mpl.args.wireless)
        else:
            return False
        return True

    def add_broadcast_mpl(self, mpl):
        mpl_in_link = {'type': 'unilink',
                       'inputs': mpl['members'],
                       'id': mpl['id'] + '_input',
                       'output': mpl, 'args': mpl['args']}

        mpl_name = mpl['id']
        self.add_node(mpl_name, 'mpl',
                     mpl)

        self.add_unilink(mpl_in_link)

        for member in mpl['members']:
            mpl_out = self.find_io_from_io_point(member, 'output')
            self.g.add_edge(mpl_name, mpl_out)

        return True

    def add_switched_mpl(self, mpl):
        mpl_name = mpl.id

        self.add_node(mpl_name, 'mpl', mpl)

        for member in mpl.members:
            if isinstance(member, str):
                key = member
            else:
                key = member['type'] + "_" + member['id']
            if not isinstance(mpl.inputs[key], str):
                self.add_unilink(mpl.inputs[key])
            if not isinstance(mpl.outputs[key], str):
                self.add_unilink(mpl.outputs[key])

        for x, _ in mpl.outputs.items():
            #y = self.find_io_from_io_point(x, 'output')
            args = obj.LinkArgs(
                loss=mpl.args.loss,
                delay=mpl.args.delay,
                bandwidth=mpl.args.bandwidth
            )
            uid = mpl.id + "_" + x
            u = obj.Unilink(
                obj.Context(''),
                uid,
                [mpl.id],
                x,
                args=args,
            )
            mpl.out_order.append(uid)
            self.add_unilink(u)
            #self.g.add_edge(mpl.id, y)

        print("moo: ", mpl.out_order)

        for x, _ in mpl.inputs.items():
            y = self.find_io_from_io_point(x, 'input')
            self.g.add_edge(y, mpl.id)

    def add_wireless(self, wireless):
        w_name = 'wireless_' + wireless['id']
        self.add_node(w_name, 'wireless', wireless)
        self.roots.append(w_name)

    def draw(self, filename=None):
        pos = nx.drawing.nx_agraph.graphviz_layout(self.g, prog='dot')
        nx.draw(self.g, pos, with_labels=True)
        if not filename:
            plt.show()
        else:
            plt.savefig(filename)

    def clear_visited(self):
        for node in self.g:
            self.g.nodes[node]['visited'] = False
