# the name `id` is part of the object schema and is required for data ingestion
# shame on python for reserving a name that is so commonly needed as a
# programatic element.
# pylint: disable=redefined-builtin
from typing import Mapping, List
from dataclasses import dataclass, field

class Context:
    def __init__(self, eid):
        self.eid = eid

# Base Classes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Object:

    def __init__(
        self,
        ctx: Context,
        id: str,
        inQ: bool = False,
        outQ: bool = False,
        root: bool = False
    ):
        self.ctx = ctx
        self.id = id
        self.inQ = inQ
        self.outQ = outQ
        self.root = root

objects: Mapping[str, type] = {}

# Objects ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

@dataclass
class WirelessLinkArgs:
    collision_handler: str

@dataclass
class LinkArgs:
    loss: float = 0.0
    delay: str = "0ms"
    bandwidth: str = "1Gbps"
    use_codel: bool = True
    additional_elements: List[str] = None
    wireless: WirelessLinkArgs = None


class Unilink(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        inputs: List[str],
        output: str,
        args=None
    ):
        if args is None:
            args = LinkArgs()
        if isinstance(args, LinkArgs):
            self.args = args
        else:
            self.args = LinkArgs(**args)

        super().__init__(ctx, id)
        self.inputs = inputs
        self.output = output

objects['unilink'] = Unilink

class Link(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        left_to_right,
        right_to_left,
    ):
        super().__init__(ctx, id)

        if left_to_right['type'] == 'unilink':
            del left_to_right['type']
            self.left_to_right = Unilink(ctx, **left_to_right)
        else:
            raise "%s: unknown link type"%id

        if right_to_left['type'] == 'unilink':
            del right_to_left['type']
            self.right_to_left = Unilink(ctx, **right_to_left)
        else:
            raise "%s: unknown link type"%id

objects['link'] = Link

class Interface(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        ip: str,
        mac: str,
        virtual: bool,
        inQ: bool = False,
        outQ: bool = False,
        root: bool = False
    ):
        super().__init__(ctx, id)
        self.ip = ip
        self.mac = mac
        self.virtual = virtual

    def prefix(self):
        return self.ip.split("/")[0]


objects['interface'] = Interface


class XDPFrom(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        mac: str,
        mode: str = 'skb',
        inQ: bool = True,
        outQ: bool = False,
        root: bool = False
    ):
        super().__init__(ctx, id, inQ=inQ)
        self.mode = mode
        self.mac = mac

objects['xdpfrom'] = XDPFrom

class XDPTo(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        mac: str,
        mode: str = 'skb',
        inQ: bool = False,
        outQ: bool = False,
        root: bool = False
    ):
        super().__init__(ctx, id, inQ=inQ)
        self.mode = mode
        self.mac = mac

objects['xdpto'] = XDPTo

class Host(Object):

    def __init__(self, ctx: Context, id: str, ifx: Interface):
        self.interface = ifx
        super().__init__(ctx, id)

class Arp(Object):

    def __init__(self, ctx: Context, id: str, ifx: Interface):
        self.interface = ifx
        super().__init__(ctx, id)

objects['host'] = Host

class VxlanDemux(Object):

    def __init__(self,
                 ctx: Context,
                 id: str,
                 vnis: List[int],
                 out_order: List[str] = None):

        if out_order is None:
            out_order = []
        self.vnis = vnis
        self.out_order = out_order
        super().__init__(ctx, id)

objects['vxlan_demux'] = VxlanDemux

class Queue(Object):

    def __init__(self, ctx: Context, id: str, size: int = int(1e6)):
        self.size = size
        super().__init__(ctx, id)


objects['queue'] = Queue

class Vxlan(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        vni: int,
        ifx: str,
        src_mac: str,
        src_ip: str,
        dst_mac: str,
        dst_ip: str,
        encap: bool
    ):
        self.vni = vni
        self.interface = ifx
        self.src_mac = src_mac
        self.src_ip = src_ip
        self.dst_mac = dst_mac
        self.dst_ip = dst_ip
        self.encap = encap

        super().__init__(ctx, id)

objects['vxlan'] = Vxlan

@dataclass
class WirelessNode:
    mac: str
    x: float
    y: float
    radius: float
    capacity: str

    def encode(self) -> str:
        return "%s %s %s %s %s\n" % (self.mac,
                                     self.x,
                                     self.y,
                                     self.radius,
                                     self.capacity)

@dataclass
class Boundaries:
    width: float
    height: float

    def encode(self) -> str:
        return "%s %s" % (self.width, self.height)

@dataclass
class WirelessTopology:
    id: str
    boundaries: Boundaries
    nodes: List[WirelessNode]


@dataclass
class Random:
    id: str
    move_probability: float
    move_speed: float
    direction_segments: int

    def encode(self) -> str:
        return '%f %f %d' % (self.move_probability,
                             self.move_speed,
                             self.direction_segments)


MOBILITY_MAP = {
    'random': Random
}

class Wireless(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        topo,
        interval: int,
        mobility_model: str,
        mobility,
        use_mobility: bool
    ):
        self.topo = WirelessTopology(**topo)
        self.interval = interval
        self.mobility_model = mobility_model
        self.mobility = MOBILITY_MAP[mobility_model](**mobility)
        self.use_mobility = use_mobility

        super().__init__(ctx, id)

objects['wireless'] = Wireless

@dataclass
class MplArgs:

    # map interfaces to macs
    macs: Mapping[str, str] = field(default_factory=lambda: {})
    loss: float = 0.0
    delay: str = "0ms"
    bandwidth: str = "1Gbps"

class Mpl(Object):

    def __init__(
        self,
        ctx: Context,
        id: str,
        mpl_type: str,
        members: List[str],
        inputs: Mapping[str, str],
        outputs: Mapping[str, str],
        args,
        in_order: List[str] = None,
        out_order: List[str] = None
    ):
        self.mpl_type = mpl_type
        self.members = members
        self.inputs = inputs
        self.outputs = outputs
        if isinstance(args, MplArgs):
            self.args = args
        else:
            self.args = MplArgs(**args)
        if in_order:
            self.in_order = in_order
        else:
            self.in_order = []
        if out_order:
            self.out_order = out_order
        else:
            self.out_order = []



        super().__init__(ctx, id)

objects['mpl'] = Mpl


class MimeFlow:
    compute: str   # request processing time
    data: int      # total data to send
    dst: str       # IP destination
    mac: str       # mac destination
    maxpkt: int    # max packet size
    minkpt: int    # min packet size
    stop: bool     # stop emulation when all flows depleted

    def __init__(
        self,
        compute: str,
        data: int = 0,
        dst: str = "",
        mac: str = "",
        maxpkt: int = 0,
        minpkt: int = 0
    ):
        self.compute = compute
        self.data = data
        self.dst = dst
        self.mac = mac
        self.maxpkt = maxpkt
        self.minpkt = minpkt


class Mime(Object):
    ''' A Mime is an element that emulates local computation and communication
    with other Mimes. A set of Mimes can be used to emulate a distributed
    workload over an emulated network. '''

    def __init__(
        self,
        ctx: Context,
        id: str,
        ip: str,
        mac: str,
        flows,
        stop: bool = False
    ):
        self.ip = ip
        self.mac = mac
        self.flows: List[MimeFlow] = [MimeFlow(**f) for f in flows]
        self.stop = stop

        super().__init__(ctx, id, inQ=True, root=True)

objects['mime'] = Mime

class Route:

    def __init__(self, prefix: str, port: int):
        self.prefix = prefix
        self.port = port

    def arg(self) -> str:
        return "%s %d"%(self.prefix, self.port)

class Router(Object):
    ''' A standard router. Supports multipath through simple round robin '''

    def __init__(
        self,
        ctx: Context,
        id: str,
        inputs: Mapping[str, str],
        outputs: Mapping[str, str],
        routes: List[Route] = None,
        in_order: List[str] = None,
        out_order: List[str] = None
    ):
        self.inputs = inputs
        self.outputs = outputs
        if routes is None:
            self.routes = []
        else:
            self.routes = [Route(**r) for r in routes]
        if in_order:
            self.in_order = in_order
        else:
            self.in_order = []
        if out_order:
            self.out_order = out_order
        else:
            self.out_order = []

        super().__init__(ctx, id)

objects['router'] = Router
