from abc import ABC, abstractmethod
import networkx as nx
from .common import find_input_port
from .element import Element

class Pipeline(ABC):

    @abstractmethod
    def build(self, g: nx.Graph, node: str, elements: Element, terminate=False):
        pass

    def encode(
        self,
        element1: str,
        element2: str,
        out_port="",
        in_port="",
        terminate=False):

        if out_port != "":
            out_port = "[%s]" % out_port

        if in_port != "":
            in_port = "[%s]" % in_port

        pipeline = element1.name + out_port + ' -> '  + in_port + element2.name

        if terminate:
            pipeline = self.terminate(pipeline)

        return pipeline


    def extend(self, pipeline: str, element: str, out_port="", in_port=""):
        if out_port != "":
            out_port = "[%s]" % out_port
        if in_port != "":
            in_port = "[%s]" % in_port

        return pipeline + out_port + ' -> ' + in_port + element.name

    def terminate(self, pipeline: str):

        return pipeline + ';'


class Simple(Pipeline):
    ''' Simple Pipelines are for nodes with 1 element and 1 output '''

    def build(
        self,
        g: nx.Graph,
        node: str,
        elements: Element,
        terminate=False
    ):

        if g.out_degree(node) != 1:
            # this may be a problem!  Nodes should only have 1 output
            return []

        # this loop is superflous as we should only have 1 neighbor but whatever
        for neighbor in g.neighbors(node):
            n_node = g.nodes[neighbor]

            if 'skip' in n_node:
                # Skip through the next node, unless you can't
                neigh_list = list(g.neighbors(neighbor))
                if neigh_list:
                    new_neigh = neigh_list[0]
                    n_node = g.nodes[new_neigh]

            # Dest nodes should have elements in sorted order
            g_node = g.nodes[node]
            our_element = g_node['elements'][0]
            next_element = n_node['elements'][0]

            in_port = ""
            if hasattr(n_node['info'], 'in_order') and node in n_node['info'].in_order:
                in_port = "%d"%n_node['info'].in_order.index(node)

            pipeline = self.encode(
                our_element,
                next_element,
                in_port=in_port,
                terminate=terminate
            )

        return [pipeline]

class Arp(Pipeline):

    def build(
        self,
        g: nx.Graph,
        node: str,
        elements: Element,
        terminate=False
    ):
        # Check correctness here! TODO
        # Make basic pipeline
        # Create classification -> ARP_RESPONDER -> output
        g_node = g.nodes[node]

        to_vxlan_pipeline = ""

        for neighbor in g.neighbors(node):
            n_node = g.nodes[neighbor]
            our_element = None

            if 'skip' in n_node:
                # Skip through the next node, unless you can't
                neigh_list = list(g.neighbors(neighbor))
                if neigh_list:
                    new_neigh = neigh_list[0]
                    n_node = g.nodes[new_neigh]

            next_element = n_node['elements'][0]

            if n_node['type'] == 'interface':
                arp_chain_pipeline = self.extend(arp_chain_pipeline, next_element)

            elif n_node['type'] == 'vxlan_demux':
                our_element = g_node['elements'][0]
                to_vxlan_pipeline = self.encode(
                    our_element,
                    next_element,
                    out_port="1",
                    terminate=terminate
                )

            else:
                # Error, unhandled linkage
                return None

        arp_chain_pipeline = self.terminate(arp_chain_pipeline)
        to_vxlan_pipeline = self.terminate(to_vxlan_pipeline)

        return [arp_chain_pipeline, to_vxlan_pipeline]

class Demux(Pipeline):
    ''' Demux pipelines are 1 to N splitters.  The graph_node needs
        to have an order object to specify which outputs go where '''

    def build(
        self,
        g: nx.Graph,
        node: str,
        elements: Element,
        terminate=False
    ):

        g_node = g.nodes[node]

        pipelines = []

        # Our 1:N element
        our_element = g_node['elements'][0]

        if g.out_degree(node) == 0:
            # this is wrong
            raise Exception("node %s out degree 0"%node)

        order = g_node['info'].out_order
        if order is None:
            # Needed an order
            raise Exception("node %s has no out_order"%node)

        print("oo: ", order)

        elements_scene = {}

        for neighbor in g.neighbors(node):
            n_node = g.nodes[neighbor]

            if 'skip' in n_node:
                # Skip through the next node, unless you can't
                neigh_list = list(g.neighbors(neighbor))
                if neigh_list:
                    neighbor = neigh_list[0]
                    n_node = g.nodes[neighbor]

            next_element = n_node['elements'][0]
            offset = 0
            if next_element.name in elements_scene:
                offset = elements_scene[next_element.name] + 1
            port = order[offset:].index(neighbor) + offset
            elements_scene[next_element.name] = port

            pipeline = self.encode(
                our_element,
                next_element,
                out_port=port,
                terminate=terminate
            )
            #pipeline = self.terminate(pipeline)
            pipelines.append(pipeline)

        return pipelines

class Chain(Pipeline):
    ''' A pipeline where all internal elements are chained
        and the last element is linked to the neighbor node element.
        Order is maintained in the element list.  Single input
        next element assumed '''

    def build(
        self,
        g: nx.Graph,
        node: str,
        elements: Element,
        terminate=False
    ):

        g_node = g.nodes[node]
        elements = g_node['elements']

        pipeline = ""
        if len(elements) < 2:
            # this cannot be a chain pipeline with < 2 elements
            return None

        pipeline = self.encode(elements[0], elements[1])

        for element in elements[2:]:
            pipeline = self.extend(pipeline, element)

        # this loop is superflous as we should only have 1 neighbor but whatever
        for neighbor in g.neighbors(node):
            n_node = g.nodes[neighbor]

            if 'skip' in n_node:
                # Skip through the next node, unless you can't
                neigh_list = list(g.neighbors(neighbor))
                if neigh_list:
                    new_neigh = neigh_list[0]
                    n_node = g.nodes[new_neigh]
                    neighbor = new_neigh
            next_element = n_node['elements'][0]
            #next_element = elements[0]
            port = find_input_port(g, node, neighbor, make_order=True) ##TODO
            if port == 0:
                pipeline = self.extend(pipeline, next_element)
            else:
                pipeline = self.extend(pipeline, next_element, in_port=port)


        if terminate:
            pipeline = self.terminate(pipeline)

        return [pipeline]
